@@ -2,9 +2,29 @@
 #ifndef _FBUFFER_H_
 #define _FBUFFER_H_
 
-#include <assert.h>
 #include "ruby.h"
 
+#ifndef RHASH_SIZE
+#define RHASH_SIZE(hsh) (RHASH(hsh)->tbl->num_entries)
+#endif
+
+#ifndef RFLOAT_VALUE
+#define RFLOAT_VALUE(val) (RFLOAT(val)->value)
+#endif
+
+#ifndef RARRAY_PTR
+#define RARRAY_PTR(ARRAY) RARRAY(ARRAY)->ptr
+#endif
+#ifndef RARRAY_LEN
+#define RARRAY_LEN(ARRAY) RARRAY(ARRAY)->len
+#endif
+#ifndef RSTRING_PTR
+#define RSTRING_PTR(string) RSTRING(string)->ptr
+#endif
+#ifndef RSTRING_LEN
+#define RSTRING_LEN(string) RSTRING(string)->len
+#endif
+
 #ifdef HAVE_RUBY_ENCODING_H
 #include "ruby/encoding.h"
 #define FORCE_UTF8(obj) rb_enc_associate((obj), rb_utf8_encoding())
@@ -145,11 +165,8 @@ static FBuffer *fbuffer_dup(FBuffer *fb)
     unsigned long len = fb->len;
     FBuffer *result;
 
-    assert(len > 0);
-    if (len > 0) {
-        result = fbuffer_alloc(len);
-        fbuffer_append(result, FBUFFER_PAIR(fb));
-    }
+    result = fbuffer_alloc(len);
+    fbuffer_append(result, FBUFFER_PAIR(fb));
     return result;
 }
 
