@@ -5,9 +5,10 @@ unless $CFLAGS.gsub!(/ -O[\dsz]?/, ' -O3')
 end
 if CONFIG['CC'] =~ /gcc/
   $CFLAGS << ' -Wall'
-  #unless $CFLAGS.gsub!(/ -O[\dsz]?/, ' -O0 -ggdb')
-  #  $CFLAGS << ' -O0 -ggdb'
-  #end
+  unless $DEBUG && !$CFLAGS.gsub!(/ -O[\dsz]?/, ' -O0 -ggdb')
+    $CFLAGS << ' -O0 -ggdb'
+  end
 end
 
+$defs << "-DJSON_GENERATOR"
 create_makefile 'json/ext/generator'
